#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_LIBPNG
#include <png.h>
#endif

#ifdef HAVE_IMLIB
# include <Imlib.h>
#endif
#ifdef HAVE_IMLIB2
# include <X11/Xlib.h>
# include <Imlib2.h>
#endif

#include "../include/fio.h"
#include "../include/tga.h"
#include "imgio.h"


#ifdef PNG_H
/* The png_jmpbuf() macro, used in error handling, became available in
 * libpng version 1.0.6.  If you want to be able to run your code with older
 * versions of libpng, you must define the macro yourself (but only if it
 * is not already defined by libpng!).
 */
#ifndef png_jmpbuf
# define png_jmpbuf(png_ptr)	((png_ptr)->jmpbuf)
#endif

/* PNG library callback data. */
typedef struct {

	char *filename;
	FILE *fp;

	png_structp png_ptr;
	png_infop png_info_ptr;
        int	png_bpl, png_bpp,	/* Geometry of PNG data. */
		png_width, png_height;
	int png_color_type;		/* One of PNG_COLOR_TYPE_*. */
	int png_last_row;
	png_bytep png_last_row_ptr;	/* Pointer to last row, used for
					 * interlaced loading.
					 */
	int png_encountered_rows;	/* Highest number of rows
					 * encountered, think of it as the
					 * *current* allocated height.
					 */

	/* PNG read buffer, contains data from the png image in
	 * multiple rows (total number of rows indicated by png_height) 
	 * and each row is allocated to the size of
	 * MAX(png_width * 4, png_bpl) (whichever is bigger).
	 *
	 * This pointer array and each row must be deallocated after use.
	 */
	png_bytep *png_data;

	/* Our allocated image buffer, size is calculated with
	 * data_bpl * png_encountered_rows.
	 */
	u_int8_t *data;
	int data_bpl, data_bpp;		/* Geometry of member data,
					 * the width and height matches
					 * png_width and png_height.
					 */
	int interlace_type;		/* One of PNG_INTERLACE_*. */
	int interlace_pass;		/* Interlace pass number. */
	int interlace_passes;		/* Total interlace passes. */

	/* Transparent color (4 bytes RGBA, only first 3 bytes are used
	 * though). If it is NULL then it hints that there is no
	 * transparency. This needs to be deallocated if it is not NULL.
	 */
	u_int8_t *trans_color;

	u_int8_t def_alpha_value;
	u_int8_t bg_color[4];		/* Background color in RGBA format. */
	int operation_completed;

	/* Progress callback. */
        void *client_data;
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        );

} imgio_png_cb_struct;
#endif	/* PNG_H */


#ifdef HAVE_IMLIB
/* Handle used for passing to all Imlib functions, not needed for
 * Imlib2. This is foundationally declared in main.c.
 */
extern void *imlib_handle;
#endif  /* HAVE_IMLIB */


/*
 *	IO type codes, these define which library to use to load or
 *	save the image. Specifying IO_TYPE_IMLIB for example will
 *	mean to use Imlib or Imlib2.
 */
#define IO_TYPE_ANY		0	/* Let us decide. */
#define IO_TYPE_IMLIB		1
#define IO_TYPE_PNG		2
#define IO_TYPE_TGA		3


/* Pointer to last load and write error message. */
static const char *last_load_error = NULL;
static const char *last_write_error = NULL;


#ifdef PNG_H
/* PNG library reading. */
static void ImgPNGReadErrorCB(png_structp png_ptr, png_const_charp mesg);
static void ImgPNGReadWarningCB(png_structp png_ptr, png_const_charp mesg);
static void ImgPNGReadConvertBuffer(
	int png_color_type,
	int width,                      /* For target and also source. */
        int height,			/* Total rows in tar_data. */
	u_int8_t *tar_data,
	int tar_bpl, int tar_bpp,
	const u_int8_t *src_data,
	int src_bpl, int src_bpp,
	int src_row_num,		/* Source row number, must be < height. */
	u_int8_t *trans_color,		/* 4 byte transparent color. */
        u_int8_t def_alpha_value
);
static void ImgPNGReadInfoCB(png_structp png_ptr, png_infop png_info_ptr);
static void ImgPNGReadRowCB(
	png_structp png_ptr, png_bytep new_row,
	png_uint_32 row_num, int pass
);
static void ImgPNGReadEndCB(png_structp png_ptr, png_infop png_info_ptr);
static int ImgPNGReadRGBA(
        const char *filename,
        int *width_rtn, int *height_rtn,
        int *bpl_rtn, int *bpp_rtn,
        u_int8_t **rgba_rtn,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format, will be modified. */
        char **creator_rtn, char **author_rtn, char **comments_rtn,
        u_int8_t def_alpha_value,
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
	),
        int *user_aborted
);
/* PNG library writing. */
static int ImgPNGWriteRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,
        u_int8_t *rgba,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format. */
	const char *creator, const char *author, const char *comments,
        int compression_level,  /* 0 to 9. */
        int interlaced,         /* 0 or 1. */
        int format,		/* 0 = b&w, 1 = grey, 2 = greyalpha
                                 * 3 = rgb, 4 = rgba.
                                 */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        ),
        int *user_aborted
);
#endif	/* PNG_H */

#ifdef TGA_H
/* TGA library reading */
static int ImgTGAReadRGBA(
        const char *filename,
        int *width_rtn, int *height_rtn,
        int *bpl_rtn, int *bpp_rtn,
        u_int8_t **rgba_rtn,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format, will be modified. */
        char **creator_rtn, char **author_rtn, char **comments_rtn,
        u_int8_t def_alpha_value,
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        ),
        int *user_aborted
);
/* TGA library writing. */
static int ImgTGAWriteRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
	int format,		/* 0 = greyscale, 1 = rgb, 2 = rgba */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        ),
        int *user_aborted
);
#endif	/* TGA_H */

/* Public. */
const char *ImgLoadGetError(void);
const char *ImgWriteGetError(void);
int ImgLoadFileRGBA(
        const char *filename,
        int *width_rtn, int *height_rtn,
        int *bpl_rtn, int *bpp_rtn,	/* bpp_rtn will always be 4. */
        u_int8_t **rgba_rtn,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format, will be modified. */
	char **creator_rtn, char **author_rtn, char **comments_rtn,
        u_int8_t def_alpha_value,
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
	)
);
int ImgWriteFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,	/* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        const char *comments,	/* May be NULL. */
        float quality,          /* 0.0 to 1.0. */
        int save_as_color,      /* True means to save as color. */
        u_int8_t def_alpha_value,
        int imlib_fallback,     /* 0 = none or 1 = ImageMajick & NetPBM. */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
);
int ImgWritePNGFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        const char *creator, const char *author, const char *comments,
        int compression_level,  /* 0 to 9. */
        int interlaced,         /* 0 or 1. */
	int format,		/* 0 = b&w, 1 = grey, 2 = greyalpha
                                 * 3 = rgb, 4 = rgba.
                                 */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
);
int ImgWriteTGAFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        int format,             /* 0 = greyscale, 1 = rgb, 2 = rgba */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
);



#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)             (((x) < 0) ? ((x) * -1) : (x))


#ifdef PNG_H
/*
 *	PNG read error callback, records the error message in a local
 *	static string.
 */
static void ImgPNGReadErrorCB(png_structp png_ptr, png_const_charp mesg)
{
	static char lmesg[1024];

	if(mesg != NULL)
	{
	    strncpy(lmesg, (const char *)mesg, 1024);
	    lmesg[1024 - 1] = '\0';

	    last_load_error = (const char *)lmesg;
	}
}
#endif  /* PNG_H */
#ifdef PNG_H
/*
 *	PNG read warning callback, records the warning message in a local
 *      static string.
 */
static void ImgPNGReadWarningCB(png_structp png_ptr, png_const_charp mesg)
{
	return;
}
#endif	/* PNG_H */
#ifdef PNG_H
/*
 *	Coppies the PNG row data given by src_data into a row inside
 *	the buffer tar_data. The allocation of the buffer tar_data is
 *	calculated by bpl * height.
 *
 *	Format of source buffer is expected to be in libpng's format.
 *	Format of target buffer is in rgba or rgb.
 */
static void ImgPNGReadConvertBuffer(
        int png_color_type,
	int width,                      /* For target and also source. */
        int height,                     /* Total rows in tar_data. */
        u_int8_t *tar_data,
        int tar_bpl, int tar_bpp,
        const u_int8_t *src_data,
        int src_bpl, int src_bpp,
        int src_row_num,		/* Source row number, must be < height. */
        u_int8_t *trans_color,          /* 4 byte transparent color. */
	u_int8_t def_alpha_value
)
{
	int bc, bc_min;
	int x, y = src_row_num;
	u_int8_t *tar_ptr;
	const u_int8_t *src_ptr;


	/* Allocation positive? */
	if((tar_bpl < 1) || (tar_bpp < 1) || (height < 1) ||
           (src_bpl < 1) || (src_bpp < 1)
	)
	    return;

	/* Source row number out of bounds? */
	if((y < 0) || (y >= height))
	    return;

	/* Buffers allocated? */
	if((tar_data == NULL) || (src_data == NULL))
	    return;

	/* Handle by color type. */
	switch(png_color_type)
	{
	  case PNG_COLOR_TYPE_GRAY:
	    /* Assume 1 byte per pixel. */
	    src_bpp = 1;
            if(1)
            {
		u_int8_t src_byte;


                /* Get smaller of target bytes per pixel or 3 bytes
		 * per pixel.
		 */
                bc_min = MIN(tar_bpp, 3);

                /* Begin copying line. */
                for(x = 0; x < width; x++)
                {
                    tar_ptr = &tar_data[(y * tar_bpl) + (x * tar_bpp)];
                    src_ptr = &src_data[x * src_bpp];

                    /* Copy source pixel bytes to the target pixel. */
		    src_byte = (*src_ptr);
                    for(bc = 0; bc < bc_min; bc++)
                        tar_ptr[bc] = src_byte;

                    /* Set rest of target pixel bytes to the default alpha
		     * value.
		     */
                    for(; bc < tar_bpp; bc++)
                        tar_ptr[bc] = def_alpha_value;
                }
            }
	    break;

	  case PNG_COLOR_TYPE_PALETTE:
	    /* Note that we expanded palette color to RGB triplets
	     * so for color type PNG_COLOR_TYPE_PALETTE, the src_bpp
	     * is actually 3 bpp (or 4 bpp if trans_color is not NULL,
	     * suggesting that there is transparency and an alpha channel).
	     */
	    if(trans_color != NULL)
		src_bpp = 4;
	    else
		src_bpp = 3;
	    if(1)
	    {
                /* Get smaller of the two pixel sizes. */
                bc_min = MIN(tar_bpp, src_bpp);

                /* Begin copying line. */
                for(x = 0; x < width; x++)
                {
                    tar_ptr = &tar_data[(y * tar_bpl) + (x * tar_bpp)];
                    src_ptr = &src_data[x * src_bpp];

                    /* Copy source pixel bytes to the target pixel. */
                    for(bc = 0; bc < bc_min; bc++)
                        tar_ptr[bc] = src_ptr[bc];

                    /* Set rest of target pixel bytes to def alpha value. */
                    for(; bc < tar_bpp; bc++)
                        tar_ptr[bc] = def_alpha_value;
		}
	    }
	    break;

	  case PNG_COLOR_TYPE_RGB:
	    if(src_bpp == 3)
	    {
		/* Get smaller of the two pixel sizes. */
		bc_min = MIN(tar_bpp, src_bpp);

		/* Begin copying line. */
		for(x = 0; x < width; x++)
		{
		    tar_ptr = &tar_data[(y * tar_bpl) + (x * tar_bpp)];
		    src_ptr = &src_data[x * src_bpp];

		    /* Copy source pixel bytes to the target pixel. */
		    for(bc = 0; bc < bc_min; bc++)
			tar_ptr[bc] = src_ptr[bc];

		    /* Set rest of target pixel bytes to def alpha value. */
                    for(; bc < tar_bpp; bc++)
                        tar_ptr[bc] = def_alpha_value;
		}
	    }
	    break;

#ifdef PNG_COLOR_TYPE_RGBA
	  case PNG_COLOR_TYPE_RGBA:
#else
	  case PNG_COLOR_TYPE_RGB_ALPHA:
#endif
            if(src_bpp == 4)
            {
                /* Get smaller of the two pixel sizes. */
                bc_min = MIN(tar_bpp, src_bpp);

                /* Begin copying line. */
                for(x = 0; x < width; x++)
                {
                    tar_ptr = &tar_data[(y * tar_bpl) + (x * tar_bpp)];
                    src_ptr = &src_data[x * src_bpp];

                    /* Copy source pixel bytes to the target pixel. */
                    for(bc = 0; bc < bc_min; bc++)
                        tar_ptr[bc] = src_ptr[bc];

                    /* Set rest of target pixel bytes to def alpha value. */
                    for(; bc < tar_bpp; bc++)
                        tar_ptr[bc] = def_alpha_value;
                }
            }
	    break;

#ifdef PNG_COLOR_TYPE_GA
          case PNG_COLOR_TYPE_GA:
#else
          case PNG_COLOR_TYPE_GRAY_ALPHA:
#endif
            /* Assume 2 bytes per pixel. */
            src_bpp = 2;
            if(1)
            {
                u_int8_t src_byte;


                /* Get smaller of target bytes per pixel or 3 bytes
                 * per pixel.
                 */
                bc_min = MIN(tar_bpp, 3);

                /* Begin copying line. */
                for(x = 0; x < width; x++)
                {
                    tar_ptr = &tar_data[(y * tar_bpl) + (x * tar_bpp)];
                    src_ptr = &src_data[x * src_bpp];

                    /* Copy source pixel bytes to the target pixel. */
                    src_byte = src_ptr[0];
                    for(bc = 0; bc < bc_min; bc++)
                        tar_ptr[bc] = src_byte;

                    /* Set rest of target pixel to source pixel's alpha
		     * value.
		     */
		    src_byte = src_ptr[1];
                    for(; bc < tar_bpp; bc++)
                        tar_ptr[bc] = src_byte;
                }
            }
	    break;
	}
}
#endif	/* PNG_H */
#ifdef PNG_H
/*
 *	PNG library read info callback.
 */
static void ImgPNGReadInfoCB(png_structp png_ptr, png_infop png_info_ptr)
{
	int color_type, bit_depth;
	int buf_len;
        imgio_png_cb_struct *cb_data = (imgio_png_cb_struct *)
            png_get_progressive_ptr(png_ptr);
        if(cb_data == NULL)
            return;

	/* Update PNG read instance pointer on core structure. */
	cb_data->png_ptr = png_ptr;
	cb_data->png_info_ptr = png_info_ptr;

	/* Get color type. */
	cb_data->png_color_type = color_type = ((png_info_ptr != NULL) ?
	    png_info_ptr->color_type : PNG_COLOR_TYPE_GRAY
	);
	bit_depth = ((png_info_ptr != NULL) ?
            png_info_ptr->bit_depth : 1
        );


	/* Insert transformations here. */

	/* Swap bytes of 16 bit files to least significant byte first */
/*	png_set_swap(png_ptr); */

	/* Reduce 16 bits data down to 8 bits data. */
	if(bit_depth == 16)
	    png_set_strip_16(png_ptr);

	/* swap the RGBA or GA data to ARGB or AG (or BGRA to ABGR) */
/*	png_set_swap_alpha(png_ptr); */

	/* Flip the RGB pixels to BGR (or RGBA to BGRA) */
/*
	if(color_type & PNG_COLOR_MASK_COLOR)
	    png_set_bgr(png_ptr);
 */

	/* Extract multiple pixels with bit depths of 1, 2, and 4 from
	 * a single byte into separate bytes (useful for paletted and
	 * grayscale images).
	 */
/*	png_set_packing(png_ptr); */

	/* Change the order of packed pixels to least significant bit
	 * first (not useful if you are using png_set_packing).
	 */
/*	png_set_packswap(png_ptr); */



	/* Expand paletted colors into true RGB triplets, this will
	 * make the actual bpp value to be 3 instead of the reported 1.
	 */
	if(color_type == PNG_COLOR_TYPE_PALETTE)
	    png_set_expand(png_ptr);
	/* Expand grayscale images to the full 8 bits from 1, 2, or 4
	 * bits per pixel.
	 */
	if((color_type == PNG_COLOR_TYPE_GRAY) && (bit_depth < 8))
	    png_set_expand(png_ptr);
	/* Expand paletted or RGB images with transparency to full alpha
	 * channels so the data will be available as RGBA quartets.
	 */
	if(png_get_valid(png_ptr, png_info_ptr, PNG_INFO_tRNS))
	{
	    /* This will give us an alpha channel. */
	    png_set_expand(png_ptr);

	    /* Also update transparent color, by cb_data->trans_color not
	     * being NULL implies there exists a transparent color and
	     * transparency.
	     */
	    cb_data->trans_color = (u_int8_t *)realloc(
		cb_data->trans_color,
		4 * sizeof(u_int8_t)
	    );
	    if(cb_data->trans_color != NULL)
	    {
		u_int8_t *trans_color = cb_data->trans_color;

		if(color_type == PNG_COLOR_TYPE_PALETTE)
		{
		    png_bytep trans = png_info_ptr->trans;
		    if((trans != NULL) && (png_info_ptr->num_trans > 0))
		    {
			memcpy(
			    trans_color, trans,
			    3 * sizeof(u_int8_t)
			);
			trans_color[3] = cb_data->def_alpha_value;
		    }
		}
		else
		{
		    png_color_16 *trans_values = &png_info_ptr->trans_values;
		    if(trans_values != NULL)
		    {
			trans_color[0] = (u_int8_t)(trans_values->red >> 8);
                        trans_color[1] = (u_int8_t)(trans_values->green >> 8);
                        trans_color[2] = (u_int8_t)(trans_values->blue >> 8);
                        trans_color[3] = cb_data->def_alpha_value;
		    }
		}
	    }
	}

	/* Get background color. */
	if(png_get_valid(png_ptr, png_info_ptr, PNG_INFO_bKGD))
	{
	    png_color_16p pBackground;
	    u_int8_t *bg_color = cb_data->bg_color;

	    /* It is not obvious from the libpng documentation, but this
	     * function takes a pointer to a pointer, and it always
	     * returns valid red, green and blue values, regardless of
	     * color_type.
	     */
	    png_get_bKGD(png_ptr, png_info_ptr, &pBackground);
	    /* However, it always returns the raw bKGD data, regardless of
	     * any bit-depth transformations, so check depth and adjust if
	     * necessary.
	     */
	    if(bit_depth == 16)
	    {
		bg_color[0] = (u_int8_t)(pBackground->red >> 8);
		bg_color[1] = (u_int8_t)(pBackground->green >> 8);
		bg_color[2] = (u_int8_t)(pBackground->blue >> 8);
		bg_color[3] = cb_data->def_alpha_value;
	    }
	    else if((color_type == PNG_COLOR_TYPE_GRAY) && (bit_depth < 8))
	    {
		if(bit_depth == 1)
		    bg_color[0] = bg_color[1] = bg_color[2] =
			pBackground->gray ? 255 : 0;
		else if(bit_depth == 2)
		    bg_color[0] = bg_color[1] = bg_color[2] =
			(255/3) * pBackground->gray;
		else /* bit_depth == 4 */
		    bg_color[0] = bg_color[1] = bg_color[2] =
			(255/15) * pBackground->gray;
		bg_color[3] = cb_data->def_alpha_value;
	    }
	    else
	    {
		bg_color[0] = (u_int8_t)pBackground->red;
		bg_color[1] = (u_int8_t)pBackground->green;
		bg_color[2] = (u_int8_t)pBackground->blue;
		bg_color[3] = cb_data->def_alpha_value;
	    }
	}


	/* Set interlace handling. */
	cb_data->interlace_passes = png_set_interlace_handling(png_ptr);

	/* Get interlace type. */
	cb_data->interlace_type = png_get_interlace_type(png_ptr, png_info_ptr);


	/* Get size of image and byte units. */
	cb_data->png_width = (int)png_get_image_width(
	    png_ptr, png_info_ptr
	);
	cb_data->png_height = (int)png_get_image_height(
            png_ptr, png_info_ptr
        );
	/* Update data_bpl when the png_width is obtained, note that
	 * that data_bpp is given at initialization of callback struct.
	 */
	cb_data->data_bpl = cb_data->data_bpp * cb_data->png_width;
        /* Get bytes per row of one PNG line. */
        cb_data->png_bpl = png_get_rowbytes(
            png_ptr, png_info_ptr
        );
        /* Get bytes per pixel of one PNG pixel. */
	if(color_type & PNG_COLOR_MASK_COLOR)
	{
	    if(color_type & PNG_COLOR_MASK_ALPHA)
		cb_data->png_bpp = 4;
	    else
		cb_data->png_bpp = 3;
	}
	else
	{
	    if(color_type & PNG_COLOR_MASK_ALPHA)
		cb_data->png_bpp = 2;
            else
                cb_data->png_bpp = 1;
        }


	/* Allocate buffer for reading libpng format data in
	 * ImgPNGRowCB(). Note that each row is uninitialized at 
	 * allocation.
	 */
	/* If already allocated, then free it first just in case. This
	 * situation should never occure.
	 */
	if(cb_data->png_data != NULL)
	{
	    int i;
	    for(i = 0; i < cb_data->png_height; i++)
		free(cb_data->png_data[i]);
	    free(cb_data->png_data);
	    cb_data->png_data = NULL;
	}
	/* Now allocate read buffer pointer array and each row, each row
	 * is of length 4 bytes per pixel of the width or the length
	 * specified by png_bpl (whichever is bigger).
	 */
	buf_len = MAX(cb_data->png_width * 4, cb_data->png_bpl) *
	    sizeof(png_byte);
	if((buf_len > 0) && (cb_data->png_height > 0))
	{
	    cb_data->png_data = (png_bytep *)malloc(
		cb_data->png_height * sizeof(png_bytep)
	    );
	    if(cb_data->png_data != NULL)
	    {
		int i;
		for(i = 0; i < cb_data->png_height; i++)
		    cb_data->png_data[i] = (png_bytep)malloc(buf_len);
            }
        }


	/* Allocate image data buffer for storing the png image data in
	 * our data format and eventually passed to calling function.
	 */
	buf_len = cb_data->data_bpl * cb_data->png_height;
	if(buf_len > 0)
	    cb_data->data = (u_int8_t *)realloc(
		cb_data->data, buf_len
	    );


	/* setjmp() doesn't make sense here, because we'd either have to
	 * exit(), longjmp() ourselves, or return control to libpng, which
	 * doesn't want to see us again.  By not doing anything here,
	 * libpng will instead jump to readpng2_decode_data(), which can
	 * return an error value to the main program.
	 */

	/* Transformations set, now begin reading by combining
	 * background with palette (this is required).
	 */
	png_start_read_image(png_ptr);
}
#endif  /* PNG_H */
#ifdef PNG_H
/*
 *	PNG library read row callback.
 */
static void ImgPNGReadRowCB(
        png_structp png_ptr, png_bytep new_row,
        png_uint_32 row_num, int pass
)
{
	imgio_png_cb_struct *cb_data = (imgio_png_cb_struct *)
	    png_get_progressive_ptr(png_ptr);
	int png_last_row;
	png_bytep png_last_row_ptr;
	if(cb_data == NULL)
	    return;

        /* Update PNG read instance pointer on core structure. */
        cb_data->png_ptr = png_ptr;

	/* Update interlace pass number. */
	cb_data->interlace_pass = pass;

	/* Get previous row (might be NULL). */
	png_last_row = cb_data->png_last_row;
	png_last_row_ptr = cb_data->png_last_row_ptr;


        /* Increment total number of rows encountered? */
        if((int)row_num >= cb_data->png_encountered_rows)
            cb_data->png_encountered_rows = MAX((int)row_num + 1, 0);

        /* Combine rows. */
        if((cb_data->png_data != NULL) && (new_row != NULL))
	{
	    int i = (int)row_num;
	    if((i >= 0) && (i < cb_data->png_height))
	    {
		png_bytep old_row = cb_data->png_data[i];
		if(old_row != NULL)
		{
		    png_progressive_combine_row(
			png_ptr, old_row, new_row
		    );

		    /* Update buffer's row that corresponds with row_num
		     * with the new_row data.
		     */
		    ImgPNGReadConvertBuffer(
			cb_data->png_color_type,
			cb_data->png_width,		/* Tar and src width. */
			cb_data->png_encountered_rows,	/* Tar height. */
			cb_data->data,
			cb_data->data_bpl, cb_data->data_bpp,
			(const u_int8_t *)old_row,
			cb_data->png_bpl, cb_data->png_bpp,
			i,				/* Row number. */
			cb_data->trans_color,
			cb_data->def_alpha_value
		    );
		}
	    }
	}

	/* Update last row information with the current row index
	 * and pointer.
	 */
	cb_data->png_last_row = (int)row_num;
	cb_data->png_last_row_ptr = new_row;
}
#endif  /* PNG_H */
#ifdef PNG_H
/*
 *	PNG library read end callback.
 */
static void ImgPNGReadEndCB(png_structp png_ptr, png_infop png_info_ptr)
{
        imgio_png_cb_struct *cb_data = (imgio_png_cb_struct *)
            png_get_progressive_ptr(png_ptr);
        if(cb_data == NULL)
            return;

        /* Update PNG read instance pointer on core structure. */
        cb_data->png_ptr = png_ptr;
        cb_data->png_info_ptr = png_info_ptr;

	/* Mark PNG reading as done. */
	cb_data->operation_completed = 1;
}
#endif  /* PNG_H */
#ifdef PNG_H
/*
 *	PNG library read front end for RGBA (4 bpp) image data.
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value (invalid format/not a png file)
 *      -3      Systems error
 *      -4      User abort
 */
static int ImgPNGReadRGBA(
	const char *filename,
        int *width_rtn, int *height_rtn,
        int *bpl_rtn, int *bpp_rtn,
        u_int8_t **rgba_rtn,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format, will be modified. */
        char **creator_rtn, char **author_rtn, char **comments_rtn,
        u_int8_t def_alpha_value,
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
	),
	int *user_aborted
)
{
	size_t bytes_read;
	static imgio_png_cb_struct *cb_data = NULL;
	static int rbuf_len = 256;
	static char *rbuf = NULL;


/* Macro to deallocate all allocated memory in this function intended
 * not to be passed to controlling function.
 */
#define DO_FREE_MEMORY	\
{ \
 /* Deallocate callback data. */ \
 if(cb_data != NULL) \
 { \
  /* Deallocate PNG read structure and info pair. */ \
  if(cb_data->png_ptr != NULL) \
  { \
   png_destroy_read_struct( \
    &cb_data->png_ptr, \
    (cb_data->png_info_ptr != NULL) ? &cb_data->png_info_ptr : NULL, \
    (png_infopp)NULL \
   ); \
   cb_data->png_ptr = NULL; \
   cb_data->png_info_ptr = NULL; \
  } \
 \
  /* Deallocate image data in libpng data format. */ \
  if(cb_data->png_data != NULL) \
  { \
   int i; \
   for(i = 0; i < cb_data->png_height; i++) \
    free(cb_data->png_data[i]); \
   free(cb_data->png_data); \
   cb_data->png_data = NULL; \
  } \
 \
  /* Deallocate return RGBA format image data. */ \
  free(cb_data->data); \
  cb_data->data = NULL; \
 \
  /* Deallocate transparent color. */ \
  free(cb_data->trans_color); \
  cb_data->trans_color = NULL; \
 \
  /* Close PNG file. */ \
  if(cb_data->fp != NULL) \
  { \
   FClose(cb_data->fp); \
   cb_data->fp = NULL; \
  } \
 \
  /* Deallocate PNG file name. */ \
  free(cb_data->filename); \
  cb_data->filename = NULL; \
 \
  /* Deallocate callback structure itself. */ \
  free(cb_data); \
  cb_data = NULL; \
 } \
 \
 /* Deallocate file reading buffer. */ \
 free(rbuf); \
 rbuf = NULL; \
}

        /* Reset global load error string pointer. */
        last_load_error = NULL;

        /* Skip entirly if user aborted. */
        if(*user_aborted)
        {
            DO_FREE_MEMORY
            last_load_error = "User aborted operation";
            return(-4);
        }

        /* Allocate callback data structure for PNG reading. */
        cb_data = (imgio_png_cb_struct *)calloc(1, sizeof(imgio_png_cb_struct));
        if(cb_data == NULL)
        {
	    DO_FREE_MEMORY
            last_load_error = "Memory allocation error";
            return(-3);
        }
        cb_data->filename = strdup(filename);
	cb_data->fp = NULL;
	cb_data->png_ptr = NULL;
	cb_data->png_info_ptr = NULL;
        cb_data->png_width = 0;		/* Obtained in ImgPNGInfoCB(). */
        cb_data->png_height = 0;
        cb_data->png_bpl = 0;
        cb_data->png_bpp = 0;
        cb_data->png_color_type = PNG_COLOR_TYPE_GRAY;
        cb_data->png_last_row_ptr = NULL;
        cb_data->png_encountered_rows = 0;
        cb_data->png_data = NULL;	/* Allocated in ImgPNGInfoCB(). */
        cb_data->data = NULL;		/* Allocated in ImgPNGInfoCB(). */
        cb_data->data_bpl = 0;		/* Calculate later. */
        cb_data->data_bpp = 4;		/* Start off with 4 bytes, RGBA. */
        cb_data->interlace_type = PNG_INTERLACE_NONE;
        cb_data->interlace_pass = 0;
        cb_data->interlace_passes = 0;
        cb_data->trans_color = NULL;	/* Allocated in ImgPNGInfoCB(). */
        cb_data->def_alpha_value = def_alpha_value;
        if(bg_color != NULL)
            memcpy(cb_data->bg_color, (const void *)bg_color, 4 * sizeof(u_int8_t));
        cb_data->operation_completed = 0;
	cb_data->client_data = client_data;
	cb_data->progress_cb = progress_cb;


	/* Open PNG file for reading in binary mode. */
	cb_data->fp = FOpen(cb_data->filename, "rb");
	if(cb_data->fp == NULL)
	{
	    DO_FREE_MEMORY
	    last_load_error = "Unable to open file for reading";
	    return(-1);
	}

	/* Read first 8 bytes to check for PNG signature? */
	if(1)
	{
	    unsigned char sig_buf[8];

	    fread(sig_buf, sizeof(unsigned char), 8, cb_data->fp);
	    if(!png_check_sig(sig_buf, 8))
	    {
		DO_FREE_MEMORY
		last_load_error = "Data does not appear to be in PNG format";
		return(-2);
	    }

	    /* Rewind so libpng can get the beginning on formal read. */
	    rewind(cb_data->fp);
	}

	/* Create instance of PNG read. */
	cb_data->png_ptr = png_create_read_struct(
	    PNG_LIBPNG_VER_STRING,
	    NULL, ImgPNGReadErrorCB, ImgPNGReadWarningCB
	);
	if(cb_data->png_ptr == NULL)
	{
	    DO_FREE_MEMORY
	    last_load_error = "Unable to create PNG read structure";
	    return(-3);
	}

	/* Create info structure for paring with PNG read instance. */
	cb_data->png_info_ptr = png_create_info_struct(cb_data->png_ptr);
	if(cb_data->png_info_ptr == NULL)
	{
	    DO_FREE_MEMORY
	    last_load_error = "Unable to create PNG info structure";
	    return(-3);
	}

        /* Allocate read buffer. */
        rbuf = (char *)realloc(rbuf, rbuf_len * sizeof(char));
        if(rbuf == NULL)
        {
	    DO_FREE_MEMORY
	    last_load_error = "Memory allocation error";
            return(-3);
        }

        /* setjmp() must be called in every function that calls a PNG-reading
         * libpng function.
         */
        if(setjmp(png_jmpbuf((cb_data->png_ptr))))
        {
	    DO_FREE_MEMORY
	    if(last_load_error == NULL)
		last_load_error = "libpng failed to load image";
            return(-1);
        }

	/* Set up PNG library progressive reading callbacks for this
	 * PNG read structure.
	 */
	png_set_progressive_read_fn(
	    cb_data->png_ptr, (void *)cb_data,
	    ImgPNGReadInfoCB, ImgPNGReadRowCB, ImgPNGReadEndCB
	);

	/* Begin reading data. */
	do
	{
	    /* Read segment of bytes from file and loaded it into
	     * the buffer.
	     */
	    bytes_read = fread(
		(void *)rbuf, sizeof(char), rbuf_len, cb_data->fp
	    );
	    if((bytes_read > 0) && (bytes_read <= rbuf_len))
	    {
		/* Process this segment of the buffer, note that
		 * that we pass the addresses to the PNG pointer
		 * and info pointer pair from the callback data so
		 * that they get updated as needed.
		 */
		png_process_data(
		    cb_data->png_ptr,
		    cb_data->png_info_ptr,
		    rbuf, bytes_read
		);

		/* Call progress callback. */
		if(cb_data->progress_cb != NULL)
		{
		    /* Report by interlace type. */
		    if(cb_data->interlace_type == PNG_INTERLACE_ADAM7)
		    {
			/* Give the current interlace pass and the total
			 * number of interlace passes needed as the min
			 * and max.
			 */
			if(!cb_data->progress_cb(
                            cb_data->client_data,
			    (cb_data->interlace_pass * cb_data->png_height) +
				cb_data->png_last_row,
			    cb_data->interlace_passes * cb_data->png_height,
			    cb_data->png_width, cb_data->png_encountered_rows,
			    cb_data->data_bpl, cb_data->data_bpp,
			    cb_data->data
			))
			{
			    (*user_aborted) = 1;
			    break;
			}
		    }
		    else
		    {
                        /* No interlacing, give the current number of rows
			 * encountered (current height) and the expected
			 * height as the min and max.
                         */
                        if(!cb_data->progress_cb(
                            cb_data->client_data,
                            cb_data->png_encountered_rows, cb_data->png_height,
                            cb_data->png_width, cb_data->png_encountered_rows,
                            cb_data->data_bpl, cb_data->data_bpp,
                            cb_data->data
                        ))
                        {
                            (*user_aborted) = 1;
                            break;
                        }
		    }
		}
	    }
	} while(!cb_data->operation_completed && (bytes_read > 0));


	/* Get creator, author, and comments. */
	if(cb_data->png_info_ptr != NULL)
	{
	    png_textp text, text_ptr;
	    int i, total_text;
	    const char *keyword;

	    png_get_text(
		cb_data->png_ptr, cb_data->png_info_ptr,
		&text, &total_text
	    );
	    for(i = 0; i < total_text; i++)
	    {
		text_ptr = &text[i];

		if(text_ptr->compression != PNG_TEXT_COMPRESSION_NONE)
		    continue;

		keyword = (const char *)text_ptr->key;
		if(keyword == NULL)
		    continue;

		if((text_ptr->text != NULL) ?
		    ((*text_ptr->text) == '\0') : 1
		)
		    continue;

		/* Author. */
		if(!strcasecmp(keyword, "Author") &&
                   (author_rtn != NULL)
		)
		{
		    free(*author_rtn);
		    (*author_rtn) = strdup(text_ptr->text);
		}
                /* E-mail. */
                else if(!strcasecmp(keyword, "E-mail") &&
                        (author_rtn != NULL)
                )
                {
		    /* Set author field as e-mail only if author field
		     * has not been set yet.
		     */
		    if((*author_rtn) == NULL)
			(*author_rtn) = strdup(text_ptr->text);
                }
                /* Creator. */
                else if(!strcasecmp(keyword, "Creator") &&
                        (creator_rtn != NULL)
                )
                {
                    free(*creator_rtn);
                    (*creator_rtn) = strdup(text_ptr->text);
                }
                /* Comments. */
                else if(!strcasecmp(keyword, "Comments") &&
                        (comments_rtn != NULL)
                )
                {
                    free(*comments_rtn);
                    (*comments_rtn) = strdup(text_ptr->text);
                }
                /* Description. */
                else if(!strcasecmp(keyword, "Description") &&
                        (comments_rtn != NULL)
                )
                {
                    /* Set author field as e-mail only if author field
                     * has not been set yet.
                     */
                    if((*comments_rtn) == NULL)
                        (*comments_rtn) = strdup(text_ptr->text);
                }
	    }
	}

	/* Write up comments if there isn't one already on the return. */
#if 0
/* Comments should not be generated, any comments should be obtained
 * from image data only.
 */
	if(((*comments_rtn) == NULL) && (cb_data->png_info_ptr != NULL))
	{
	    char s[256];


	    (*s) = '\0';
	    png_info_ptr = cb_data->png_info_ptr;
	    switch(png_info_ptr->color_type)
	    {
	      case PNG_COLOR_TYPE_GRAY:
		if(cb_data->interlace_passes > 1)
		    sprintf(
			s,
"Greyscale, interlaced (%i passes)",
			cb_data->interlace_passes
		    );
		else
                    sprintf(
                        s,
"Greyscale, non-interlaced"
                    );
		break;

	      case PNG_COLOR_TYPE_PALETTE:
		if(cb_data->interlace_passes > 1)
                    sprintf(
                        s,
"Palette with %i colors, interlaced (%i passes)",
			png_info_ptr->num_palette,
                        cb_data->interlace_passes
                    );
                else
                    sprintf(
                        s,
"Palette with %i colors, non-interlaced",
			png_info_ptr->num_palette
                    );
		break;

	      case PNG_COLOR_TYPE_RGB:
                if(cb_data->interlace_passes > 1)
                    sprintf(
                        s,
"Rgb, interlaced (%i passes)",
                        cb_data->interlace_passes
                    );
                else
                    sprintf(
                        s,
"Rgb, non-interlaced"
                    );
                break;

#ifdef PNG_COLOR_TYPE_RGBA
              case PNG_COLOR_TYPE_RGBA:
#else
              case PNG_COLOR_TYPE_RGB_ALPHA:
#endif
                if(cb_data->interlace_passes > 1)
                    sprintf(
                        s,
"Rgb with alpha channel, interlaced (%i passes)",
                        cb_data->interlace_passes
                    );
                else
                    sprintf(
                        s,
"Rgb with alpha channel, non-interlaced"
                    );
                break;

#ifdef PNG_COLOR_TYPE_GA
	      case PNG_COLOR_TYPE_GA:
#else
	      case PNG_COLOR_TYPE_GRAY_ALPHA:
#endif
                if(cb_data->interlace_passes > 1)
                    sprintf(
                        s,
"Greyscale with alpha channel, interlaced (%i passes)",
                        cb_data->interlace_passes
                    );
                else
                    sprintf(
                        s,
"Greyscale with alpha channel, non-interlaced"
                    );
		break;
	    }
	    /* If comments string was properly formatted, then copy it to
	     * the comments return.
	     */
	    if((*s) != '\0')
		(*comments_rtn) = strdup(s);


	    /* Needs to be NULL when we return control from this case. */
	    png_info_ptr = NULL;
	}
#endif

	/* Update returns. */
	if(width_rtn != NULL)
	    (*width_rtn) = cb_data->png_width;
        if(height_rtn != NULL)
            (*height_rtn) = cb_data->png_height;
        if(bpl_rtn != NULL)
            (*bpl_rtn) = cb_data->data_bpl;
        if(bpp_rtn != NULL)
            (*bpp_rtn) = cb_data->data_bpp;

        if(rgba_rtn != NULL)
	{
	    /* Transfer rgba data. */
	    free(*rgba_rtn);
            (*rgba_rtn) = cb_data->data;

	    /* Mark it NULL so it dosen't get deallocated when callback
	     * data is deallocated.
	     */
	    cb_data->data = NULL;
	}

        if(bg_color != NULL)
            memcpy(bg_color, cb_data->bg_color, 4 * sizeof(u_int8_t));


	/* Deallocate callback structure with all its resources,
	 * destroy PNG read instance and info pointer pair from the
         * callback data structure, close the PNG file, and deallocate 
	 * the file read buffer.
         */
	DO_FREE_MEMORY
#undef DO_FREE_MEMORY

	if(*user_aborted)
	{
	    last_load_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(0);
}
#endif  /* PNG_H */
#ifdef PNG_H
/*
 *      PNG library write front end for RGBA (4 bpp) image data.
 *
 *	Format means the format to be written to the PNG file, not the
 *	format of the given rgba data.
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value (invalid format)
 *      -3      Systems error
 *      -4      User abort
 */
static int ImgPNGWriteRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
	const char *creator, const char *author, const char *comments,
	int compression_level,	/* 0 to 9. */
	int interlaced,		/* 0 or 1. */
	int format,		/* 0 = b&w, 1 = grey, 2 = greyalpha
				 * 3 = rgb, 4 = rgba.
				 */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        ),
        int *user_aborted
)
{
	static imgio_png_cb_struct *cb_data;


/* Macro to deallocate all allocated memory in this function intended
 * not to be passed to controlling function.
 */
#define DO_FREE_MEMORY	\
{ \
 /* Deallocate callback data. */ \
 if(cb_data != NULL) \
 { \
  /* Deallocate PNG write structure and info pair. */ \
  if(cb_data->png_ptr != NULL) \
  { \
   png_destroy_write_struct( \
    &cb_data->png_ptr, \
    (cb_data->png_info_ptr != NULL) ? &cb_data->png_info_ptr : NULL \
   ); \
   cb_data->png_ptr = NULL; \
   cb_data->png_info_ptr = NULL; \
  } \
 \
  /* Deallocate image data in libpng data format. */ \
  if(cb_data->png_data != NULL) \
  { \
   int i; \
   for(i = 0; i < cb_data->png_height; i++) \
    free(cb_data->png_data[i]); \
   free(cb_data->png_data); \
   cb_data->png_data = NULL; \
  } \
 \
  /* Deallocate return RGBA format image data. */ \
  free(cb_data->data); \
  cb_data->data = NULL; \
 \
  /* Deallocate transparent color. */ \
  free(cb_data->trans_color); \
  cb_data->trans_color = NULL; \
 \
  /* Close PNG file. */ \
  if(cb_data->fp != NULL) \
  { \
   FClose(cb_data->fp); \
   cb_data->fp = NULL; \
  } \
 \
  /* Deallocate PNG file name. */ \
  free(cb_data->filename); \
  cb_data->filename = NULL; \
 \
  /* Deallocate callback structure itself. */ \
  free(cb_data); \
  cb_data = NULL; \
 } \
}


        /* Reset global write error string pointer. */
        last_write_error = NULL;

        /* Skip entirly if user aborted. */
        if(*user_aborted)
        {
	    DO_FREE_MEMORY
            last_write_error = "User aborted operation";
            return(-4);
        }

	if((rgba == NULL) || (width < 1) || (height < 1) ||
           (bpp != 4)
	)
	{
	    DO_FREE_MEMORY
	    last_write_error = "Invalid description of image data";
	    return(-1);
	}

        /* Automatically calculate bytes per line? */
        if(bpl <= 0)
            bpl = width * bpp;
        if(bpl < 1)
        {
	    DO_FREE_MEMORY
            last_write_error = "Invalid description of image data";
            return(-1);
        }


        /* Allocate callback data structure for PNG writing. */
        cb_data = (imgio_png_cb_struct *)calloc(1, sizeof(imgio_png_cb_struct));
        if(cb_data == NULL)
        {
            DO_FREE_MEMORY
            last_load_error = "Memory allocation error";
            return(-3);
        }
        cb_data->filename = strdup(filename);
        cb_data->fp = NULL;
        cb_data->png_ptr = NULL;
        cb_data->png_info_ptr = NULL;
        cb_data->png_width = width;
        cb_data->png_height = height;
        cb_data->png_bpl = 0;		/* Not used. */
        cb_data->png_bpp = 0;		/* Not used. */
        cb_data->png_color_type = PNG_COLOR_TYPE_GRAY;
        cb_data->png_last_row_ptr = NULL;
        cb_data->png_encountered_rows = 0;
        cb_data->png_data = NULL;	/* Allocated later. */
        cb_data->data = NULL;		/* Never used or allocated. */
        cb_data->data_bpl = bpl;
        cb_data->data_bpp = bpp;	/* Must be 4 bytes RGBA. */
        cb_data->interlace_type = PNG_INTERLACE_NONE;
        cb_data->interlace_pass = 0;
        cb_data->interlace_passes = 0;
        cb_data->trans_color = NULL;    /* Allocated in ImgPNGInfoCB(). */
        cb_data->def_alpha_value = 0xff;	/* Not used, so set it anything. */
        if(bg_color != NULL)
            memcpy(cb_data->bg_color, (const void *)bg_color, 4 * sizeof(u_int8_t));
        cb_data->operation_completed = 0;
        cb_data->client_data = client_data;
        cb_data->progress_cb = progress_cb;


        /* Set color_type depending on the given format, format can be:
         *
         *      0 = b&w
         *      1 = Greyscale
         *      2 = Greyscale with alpha
         *      3 = RGB
         *      4 = RGBA
         */
        switch(format)
        {
          case 0:
          case 1:
	    cb_data->png_color_type = PNG_COLOR_TYPE_GRAY;
            break;
          case 2:
#ifdef PNG_COLOR_TYPE_GA
            cb_data->png_color_type = PNG_COLOR_TYPE_GA;
#else
            cb_data->png_color_type = PNG_COLOR_TYPE_GRAY_ALPHA;
#endif
            break;
          case 3:
	    cb_data->png_color_type = PNG_COLOR_TYPE_RGB;
            break;
          case 4:
#ifdef PNG_COLOR_TYPE_RGBA
	    cb_data->png_color_type = PNG_COLOR_TYPE_RGBA;
#else
	    cb_data->png_color_type = PNG_COLOR_TYPE_RGB_ALPHA;
#endif
            break;

          default:
	    cb_data->png_color_type = PNG_COLOR_TYPE_GRAY;
            break;
        }

        /* Set interlace type. */
        if(interlaced)
	    cb_data->interlace_type = PNG_INTERLACE_ADAM7;
        else
	    cb_data->interlace_type = PNG_INTERLACE_NONE;

        /* Allocate libpng format image data in multiple rows. */
	cb_data->png_data = (png_bytep *)realloc(
            cb_data->png_data, cb_data->png_height * sizeof(png_bytep)
        );
        if(cb_data->png_data != NULL)
        {
            int x, y, png_row_len;
            png_bytep png_row_ptr;
            const u_int8_t *src_ptr;
	    png_bytep *png_data = cb_data->png_data;


            /* Set color_type depending on the given format, format can be:
             *
             *  0 = b&w
             *  1 = Greyscale
             *  2 = Greyscale with alpha
             *  3 = RGB
             *  4 = RGBA
             */
            switch(format)
            {
              case 0:
              case 1:
                /* Greyscale. */
                png_row_len = 1 * cb_data->png_width * sizeof(png_byte);
                for(y = 0; y < cb_data->png_height; y++)
                {
                    png_data[y] = png_row_ptr = (png_bytep)malloc(png_row_len);
                    if(png_row_ptr == NULL)
                        continue;

                    if(cb_data->progress_cb != NULL)
                    {
                        /* Report progress from 0 to 50%. */
                        if(!cb_data->progress_cb(
                            cb_data->client_data,
                            (int)(y / 2.0), cb_data->png_height,
                            cb_data->png_width, cb_data->png_height,
                            cb_data->data_bpl, cb_data->data_bpp,
                            rgba
                        ))
                        {
                            (*user_aborted) = 1;
                            break;
                        }
                    }

                    for(x = 0; x < cb_data->png_width; x++)
                    {
                        src_ptr = (const u_int8_t *)&rgba[
                            (y * cb_data->data_bpl) + (x * cb_data->data_bpp)
                        ];
                        *png_row_ptr++ = (png_byte)(
                            (0.299 * src_ptr[0]) + (0.587 * src_ptr[1]) +
                            (0.114 * src_ptr[2])
                        );
                    }
                }
                break;

              case 2:
                /* Greyscale Alpha. */
                png_row_len = 2 * cb_data->png_width * sizeof(png_byte);
                for(y = 0; y < cb_data->png_height; y++)
                {
                    png_data[y] = png_row_ptr = (png_bytep)malloc(png_row_len);
                    if(png_row_ptr == NULL)
                        continue;

                    if(cb_data->progress_cb != NULL)
                    {
                        /* Report progress from 0 to 50%. */
                        if(!cb_data->progress_cb(
                            cb_data->client_data,
                            (int)(y / 2.0), cb_data->png_height,
                            cb_data->png_width, cb_data->png_height,
                            cb_data->data_bpl, cb_data->data_bpp,
                            rgba
                        ))
                        {
                            (*user_aborted) = 1;
                            break;
                        }
                    }

                    for(x = 0; x < cb_data->png_width; x++)
                    {
                        src_ptr = (const u_int8_t *)&rgba[
                            (y * cb_data->data_bpl) + (x * cb_data->data_bpp)
                        ];
                        *png_row_ptr++ = (png_byte)(
                            (0.299 * src_ptr[0]) + (0.587 * src_ptr[1]) +
                            (0.114 * src_ptr[2])
                        );
                        *png_row_ptr++ = (png_byte)src_ptr[3];
                    }
                }
                break;

              case 3:
                /* RGB. */
                png_row_len = 3 * cb_data->png_width * sizeof(png_byte);
                for(y = 0; y < cb_data->png_height; y++)
                {
                    png_data[y] = png_row_ptr = (png_bytep)malloc(png_row_len);
                    if(png_row_ptr == NULL)
                        continue;

                    if(progress_cb != NULL)
                    {
                        /* Report progress from 0 to 50%. */
                        if(!cb_data->progress_cb(
                            cb_data->client_data,
                            (int)(y / 2.0), cb_data->png_height,
                            cb_data->png_width, cb_data->png_height,
                            cb_data->data_bpl, cb_data->data_bpp,
                            rgba
                        ))
                        {
                            (*user_aborted) = 1;
                            break;
                        }
                    }

                    for(x = 0; x < cb_data->png_width; x++)
                    {
                        src_ptr = (const u_int8_t *)&rgba[
                            (y * cb_data->data_bpl) + (x * cb_data->data_bpp)
                        ];
                        *png_row_ptr++ = src_ptr[0];
                        *png_row_ptr++ = src_ptr[1];
                        *png_row_ptr++ = src_ptr[2];
                    }
                }
                break;

              case 4:
                /* RGBA. */
                png_row_len = 4 * cb_data->png_width * sizeof(png_byte);
                for(y = 0; y < cb_data->png_height; y++)
                {
                    png_data[y] = png_row_ptr = (png_bytep)malloc(png_row_len);
                    if(png_row_ptr == NULL)
                        continue;

                    if(cb_data->progress_cb != NULL)
                    {
                        /* Report progress from 0 to 50%. */
                        if(!cb_data->progress_cb(
                            cb_data->client_data,
                            (int)(y / 2.0), cb_data->png_height,
                            cb_data->png_width, cb_data->png_height,
                            cb_data->data_bpl, cb_data->data_bpp,
                            rgba
                        ))
                        {
                            (*user_aborted) = 1;
                            break;
                        }
                    }
                    for(x = 0; x < cb_data->png_width; x++)
                    {
                        src_ptr = (const u_int8_t *)&rgba[
                            (y * cb_data->data_bpl) + (x * cb_data->data_bpp)
                        ];
                        *png_row_ptr++ = src_ptr[0];
                        *png_row_ptr++ = src_ptr[1];
                        *png_row_ptr++ = src_ptr[2];
                        *png_row_ptr++ = src_ptr[3];
                    }
                }
                break;
            }
	}


        /* Open PNG file for writing in binary mode. */
	cb_data->fp = FOpen(cb_data->filename, "wb");
        if(cb_data->fp == NULL)
        {
	    DO_FREE_MEMORY
            last_write_error = "Unable to open file for writing";
            return(-1);
        }

	/* Create PNG write structure. */
	cb_data->png_ptr = png_create_write_struct(
	    PNG_LIBPNG_VER_STRING, NULL, NULL, NULL
	);
	if(cb_data->png_ptr == NULL)
	{
	    DO_FREE_MEMORY
	    last_write_error = "Unable to create PNG write structure";
	    return(-3);
	}

	/* Create info structure. */
	cb_data->png_info_ptr = png_create_info_struct(cb_data->png_ptr);
	if(cb_data->png_info_ptr == NULL)
	{
	    DO_FREE_MEMORY
	    last_write_error = "Unable to create PNG info structure";
	    return(-3);
	}

	/* setjmp() must be called in every function that calls a
	 * PNG-writing libpng function.
	 */
	if(setjmp(png_jmpbuf(cb_data->png_ptr)))
	{
	    DO_FREE_MEMORY
	    if(last_write_error)
		last_write_error = "libpng failed to save image";
            return(-1);
        }

	/* Set output stream. */
	png_init_io(cb_data->png_ptr, cb_data->fp);

	/* Begin setting write parameters. */

	/* Compression level (0 to 9). */
	png_set_compression_level(
	    cb_data->png_ptr,
	    CLIP(compression_level, 0, 9)
	);

/*
this is default for no filtering; Z_FILTERED is default otherwise:
	png_set_compression_strategy(cb_data->png_ptr, Z_DEFAULT_STRATEGY);
these are all defaults:
	png_set_compression_mem_level(cb_data->png_ptr, 8);
	png_set_compression_window_bits(cb_data->png_ptr, 15);
	png_set_compression_method(cb_data->png_ptr, 8);
 */

	/* Set up image header. */
	png_set_IHDR(
	    cb_data->png_ptr, cb_data->png_info_ptr,
	    cb_data->png_width, cb_data->png_height,	/* Size. */
	    8,			/* Depth per pixel compoent. */
	    cb_data->png_color_type, cb_data->interlace_type,
	    PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT
	);
/*
	if(gamma > 0.0)
	    png_set_gAMA(png_ptr, png_info_ptr, gamma);
 */

	/* Set background color. */
	if(cb_data->bg_color != NULL)
	{
	    png_color_16 bg;

	    /* Format in 8 bit RGBA. */
	    bg.red = cb_data->bg_color[0];
            bg.green = cb_data->bg_color[1];
            bg.blue = cb_data->bg_color[2];
/*	    ??? = cb_data->bg_color[3];   skip alpha */
	    bg.gray = ((int)cb_data->bg_color[0] +
		(int)cb_data->bg_color[1] + (int)cb_data->bg_color[2]) / 3;
	    png_set_bKGD(cb_data->png_ptr, cb_data->png_info_ptr, &bg);
	}

	/* Set modified time. */
/*
	if(mainprog_ptr->have_time)
	{
	    png_time modtime;

	    png_convert_from_time_t(&modtime, mainprog_ptr->modtime);
	    png_set_tIME(cb_data->png_ptr, cb_data->png_info_ptr, &modtime);
	}
 */

	/* Write up text fields. */
	if(1)
	{
#define total_text	7
	    png_text text[total_text];
	    int num_text = 0;


	    if(0)
	    {
		text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
		text[num_text].key = "Title";
		text[num_text].text = "Untitled";
		++num_text;
	    }
            if(creator != NULL)
            {
                /* PNG dosen't have a standard keyword (or any keyword
		 * that comes close) to suggest creator, so implicitly
		 * use a keyword called "Creator".
		 */
                text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
                text[num_text].key = "Creator";
                text[num_text].text = (char *)creator;
                ++num_text;
            }
            if(author != NULL)
            {
                text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
                text[num_text].key = "Author";
                text[num_text].text = (char *)author;
                ++num_text;
            }
            if(comments != NULL)
	    {
		/* Save comments under PNG keyword "Description". */
		text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
		text[num_text].key = "Description";
		text[num_text].text = (char *)comments;
		++num_text;
	    }
            if(0)
	    {
                text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
                text[num_text].key = "Copyright";
                text[num_text].text = "By Author";
                ++num_text;
            }
            if(0)
            {
                text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
                text[num_text].key = "E-mail";
                text[num_text].text = "Undefined";
                ++num_text;
            }
            if(0)
	    {
                text[num_text].compression = PNG_TEXT_COMPRESSION_NONE;
                text[num_text].key = "URL";
                text[num_text].text = "None";
                ++num_text;
            }
            png_set_text(cb_data->png_ptr, cb_data->png_info_ptr, text, num_text);

#undef total_text
	}

	/* Write all chunks up to (but not including) first IDAT */
	png_write_info(cb_data->png_ptr, cb_data->png_info_ptr);


        /* Set up the transformations:  for now, just pack low-bit-depth
	 * pixels into bytes (one, two or four pixels per byte)
	 */
	png_set_packing(cb_data->png_ptr);
/*  png_set_shift(cb_data->png_ptr, &sig_bit);  to scale low-bit-depth values */



	/* Begin writing. */

	if((cb_data->png_data != NULL) && !(*user_aborted))
	{
	    /* Handle by interlaced or non-interlacted. */
	    if(cb_data->interlace_type == PNG_INTERLACE_NONE)
	    {
		int y;

		/* For non-interlaced, each row is written at a
		 * at a time.
		 */
		for(y = 0; y < cb_data->png_height; y++)
		{
		    if(cb_data->progress_cb != NULL)
		    {
			/* Report progress from 50% to 100%. */
			if(!cb_data->progress_cb(
			    cb_data->client_data,
			    (int)(y / 2.0) + (cb_data->png_height / 2), cb_data->png_height,
			    cb_data->png_width, cb_data->png_height,
			    cb_data->data_bpl, cb_data->data_bpp,
			    rgba
                        ))
                        {
                            (*user_aborted) = 1;
                            break;
                        }
                    }
		    /* Write this row. */
		    png_write_row(cb_data->png_ptr, cb_data->png_data[y]);
		}
	    }
	    else
	    {
		/* For interlacing the entire image must be written at
		 * once.
		 */
		png_write_image(cb_data->png_ptr, cb_data->png_data);
	    }
	}

	/* Since that's it, we also close out the end of the PNG file now.
	 * If we had any text or time info to write after the IDATs, second
	 * argument would be info_ptr, but we optimize slightly by sending NULL 
	 * pointer.
	 */
	png_write_end(cb_data->png_ptr, NULL);


        /* Deallocate callback structure with all its resources,
         * destroy PNG write instance and info pointer pair from the
         * callback data structure, and close the PNG file.
         */
	DO_FREE_MEMORY
#undef DO_FREE_MEMORY

        if(*user_aborted)
        {
            last_write_error = "User aborted operation";
            return(-4);
        }
        else
            return(0);
}
#endif  /* PNG_H */


#ifdef TGA_H
/*
 *	TGA library read front end for RGBA (4 bpp) image data.
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value (invalid format/not a tga file)
 *      -3      Systems error
 *      -4      User abort
 */
static int ImgTGAReadRGBA(
        const char *filename,
        int *width_rtn, int *height_rtn,
        int *bpl_rtn, int *bpp_rtn,
        u_int8_t **rgba_rtn,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format, will be modified. */
        char **creator_rtn, char **author_rtn, char **comments_rtn,
        u_int8_t def_alpha_value,
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        ),
        int *user_aborted
)
{
	u_int8_t *rgba;
	int width, height, bpl, bpp;

        /* Reset global load error string pointer. */
        last_load_error = NULL;

        /* Skip entirly if user aborted. */
        if(*user_aborted)
        {
            last_load_error = "User aborted operation";
            return(-4);
        }

	/* Read TGA image */
	rgba = TgaReadFromFileFastRGBA(
	    filename, &width, &height,
	    (bg_color != NULL) ? *(u_int32_t *)bg_color : 0x00000000
	);
	bpp = 4;
	bpl = width * bpp;

	if(rgba_rtn != NULL)
	    *rgba_rtn = rgba;
        if(width_rtn != NULL)
            *width_rtn = width;
        if(height_rtn != NULL)
            *height_rtn = height;
        if(bpl_rtn != NULL)
            *bpl_rtn = bpl;
        if(bpp_rtn != NULL)
            *bpp_rtn = bpp;

        if(*user_aborted)
        {
            last_load_error = "User aborted operation";
            return(-4);
        }
        else
            return((rgba != NULL) ? 0 : -1);
}
#endif	/* TGA_H */
#ifdef TGA_H
/*
 *      TGA library write front end for RGBA (4 bpp) image data.
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value (invalid format)
 *      -3      Systems error
 *      -4      User abort
 */
static int ImgTGAWriteRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
	int format,		/* 0 = greyscale, 1 = rgb, 2 = rgba */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        ),
        int *user_aborted
)
{
	int status, tga_depth;
	tga_data_struct td_data, *td = &td_data;


        /* Reset global write error string pointer. */
        last_write_error = NULL;

        /* Skip entirly if user aborted. */
        if(*user_aborted)
        {
            last_write_error = "User aborted operation";
            return(-4);
        }

        if((rgba == NULL) || (width < 1) || (height < 1) ||
           (bpp != 4)
        )
        {
            last_write_error = "Invalid description of image data";
            return(-1);
        }

	if(bpl <= 0)
	    bpl = width * bpp;

	/* Select tga destination depth */
	switch(format)
	{
	  case 1:	/* 24 bits rgb */
	    tga_depth = 24;
	    break;

          case 0:	/* 8 bits greyscale */
            tga_depth = 8;
            break;

	  default:	/* 32 bits rgba */
	    tga_depth = 32;
	    break;
	}

	/* Set up tga data for writing */
	memset(td, 0x0, sizeof(tga_data_struct));
	td->x = 0;
	td->y = 0;
	td->width = width;
	td->height = height;
	td->depth = 32;	/* Ignored */
	td->data = (u_int8_t *)malloc(bpl * height);
	td->data_depth = bpp * 8;

	/* Need to convert source RGBA to target BGRA as we copy the
	 * source RGBA to the target buffer.
	 */
	if(td->data != NULL)
	{
	    u_int8_t	*tar = td->data,
			*tar_end = tar + (bpl * height);
	    const u_int8_t *src = rgba;

	    while(tar < tar_end)
	    {
		tar[0] = src[2];
		tar[1] = src[1];
		tar[2] = src[0];
		tar[3] = src[3];
		tar += bpp;
		src += bpp;

#if 0
/* Too cpu consuming */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			tar - td->data, tar_end - td->data,
			width, height,
			bpl, bpp,
			rgba
		    ))
			*user_aborted = 1;
		}
		if(*user_aborted)
		    break;
#endif
	    }
	}

	/* Write tga image */
	if(*user_aborted)
	    status = TgaNoAccess;
	else
	    status = TgaWriteToFile(
		filename,
		td,
		tga_depth	/* Target depth */
	    );

	/* Delete tga data */
	TgaDestroyData(td);

	if(*user_aborted)
	    return(-4);
	else
	    return((status == TgaSuccess) ? 0 : -1);
}
#endif  /* TGA_H */


/*
 *	Returns the pointer to a error message string describing
 *	the last error encountered for a prior call to ImgLoad*().
 *
 *	Can return NULL if there was no error.
 */
const char *ImgLoadGetError(void)
{
	return(last_load_error);
}

/*
 *      Returns the pointer to a error message string describing
 *      the last error encountered for a prior call to ImgWrite*().
 *
 *      Can return NULL if there was no error.
 */
const char *ImgWriteGetError(void)
{
	return(last_write_error);
}


/*
 *	Loads the image from file.
 *
 *	All return values must be valid!
 *
 *	If progress_cb is not NULL then it will be called during loading,
 *	if it returns 0 then loading will be aborted.
 *
 *	Calling function must deallocate the returned rgba_rtn and
 *	comments_rtn pointers (the ones that are not returned NULL).
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Unable to determine file format
 *      -3      Systems error
 *      -4      User abort
 */
int ImgLoadFileRGBA(
        const char *filename,
        int *width_rtn, int *height_rtn,
        int *bpl_rtn, int *bpp_rtn,
        u_int8_t **rgba_rtn,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format, will be modified. */
	char **creator_rtn, char **author_rtn, char **comments_rtn,
        u_int8_t def_alpha_value,
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
	)
)
{
	int status;
	int user_aborted = 0;
	int io_type = IO_TYPE_ANY;
	const char *ext_ptr;
	int	width = 0, height = 0,
		bpl = 0, bpp = 0;
	u_int8_t *data_rgba = NULL;
	char *creator = NULL, *author = NULL, *comments = NULL;


	/* Reset global last load error message pointer. */
	last_load_error = NULL;

	/* All return values must be valid. */
	if((width_rtn == NULL) || (height_rtn == NULL) ||
           (bpl_rtn == NULL) || (bpp_rtn == NULL) ||
           (rgba_rtn == NULL)
	)
	{
            last_load_error = "Internal error: Invalid return address for image data description";
	    return(-1);
	}

	/* Reset returns. */
	*width_rtn = width;
	*height_rtn = height;
	*bpl_rtn = bpl;
	*bpp_rtn = bpp;
	*rgba_rtn = data_rgba;
	*creator_rtn = creator;
	*author_rtn = author;
	*comments_rtn = comments;

	/* File name must be given. */
        if(filename == NULL)
	{
	    last_load_error = "Internal error: File name memory address is NULL";
            return(-1);
	}

	/* Get pointer to extension (if any) in file name. */
	ext_ptr = strrchr(filename, '.');
	if(ext_ptr == NULL)
	    ext_ptr = "";

	/* Is the IO type undefined? */
	if(io_type == IO_TYPE_ANY)
	{
	    if((*ext_ptr) == '\0')
	    {
		/* No extension available, so assume IO_TYPE_IMLIB. */
		io_type = IO_TYPE_IMLIB;
	    }
	    else
	    {
		/* Determine IO type from extension. */
		if(!strcasecmp(ext_ptr, ".png"))
		    io_type = IO_TYPE_PNG;
		else if(!strcasecmp(ext_ptr, ".tga"))
		    io_type = IO_TYPE_TGA;
		else
		    io_type = IO_TYPE_IMLIB;
	    }
	}


	/* Reset status to -1 and load by IO type. */
	status = -1;

	/* Call progress callback for the first time as needed. */
	if(progress_cb != NULL)
	{
	    if(!progress_cb(
		client_data,
		0, 1,		/* Progress starting. */
                width, height,
                bpl, bpp,
		data_rgba
	    ))
		user_aborted = 1;
	}

	/* Use Imlib or Imlib2? */
	if((io_type == IO_TYPE_IMLIB) && !user_aborted)
	{
#ifdef HAVE_IMLIB
	    if(imlib_handle != NULL)
	    {
		const u_int8_t *src_data_rgb, *src_data_alpha;
		char *dpath;
		ImlibImage *imlib_image;


		/* Copy path and load Imlib image. */
		dpath = strdup(filename);
		imlib_image = Imlib_load_image(imlib_handle, dpath);
		free(dpath); dpath = NULL;

		/* Image loaded successfully? */
		if(imlib_image != NULL)
		{
		    status = 0;		/* Mark that load was successful. */

		    /* Need to realize changes. */
		    Imlib_changed_image(imlib_handle, imlib_image);

		    /* Get pointer to buffers and size of newly loaded image. */
		    src_data_rgb = (const u_int8_t *)imlib_image->rgb_data;
		    src_data_alpha = (const u_int8_t *)imlib_image->alpha_data;

		    /* Update image information. */
		    width = imlib_image->rgb_width;
		    height = imlib_image->rgb_height;
		    bpp = 4;
		    bpl = width * bpp;

		    /* Make sure we have the rgb data and a positive geometry. */
		    if((src_data_rgb != NULL) && (width > 0) && (height > 0))
		    {
			int src_bpl_rgb, src_bpl_alpha;


			/* Source image rgb buffer bytes per line. */
			src_bpl_rgb = width * (3 * sizeof(u_int8_t));
			/* Source image alpha buffer bytes per line. */
			src_bpl_alpha = width * (1 * sizeof(u_int8_t));

			/* Allocate return image data buffer. */
			data_rgba = (u_int8_t *)realloc(
			    data_rgba,
			    bpl * height
			);
			if(data_rgba != NULL)
			{
			    int x, y;
			    u_int8_t *tar_ptr;
			    const u_int8_t *src_ptr;

			    /* Itterate through each line. */
			    for(y = 0; y < height; y++)
			    {
				/* Call progress callback. */
				if(progress_cb != NULL)
				{
				    if(!progress_cb(
					client_data, y, height,
					width, height,
					bpl, bpp,
					data_rgba
				    ))
				    {
					user_aborted = 1;
					break;
				    }
				}

				/* Itterate through current line. */
				for(x = 0; x < width; x++)
				{
				    tar_ptr = &data_rgba[
					(y * bpl) + (x * bpp)
				    ];
				    src_ptr = &src_data_rgb[
					(y * src_bpl_rgb) +
					(x * (3 * sizeof(u_int8_t)))
                                    ];

                                    /* Copy rgb of this pixel (3 bytes). */
                                    *tar_ptr++ = *src_ptr++;
                                    *tar_ptr++ = *src_ptr++;
                                    *tar_ptr++ = *src_ptr;

                                    /* Alpha channel data buffer avalable? */
                                    if(src_data_alpha != NULL)
                                    {
                                        src_ptr = &src_data_alpha[
                                            (y * src_bpl_alpha) +
                                            (x * (1 * sizeof(u_int8_t)))
                                        ];
                                        /* Copy alpha of this pixel (1 byte). */
                                        *tar_ptr = *src_ptr;
                                    }
                                    else
                                    {
                                        /* Set default alpha value. */
                                        *tar_ptr = def_alpha_value;
                                    }
                                }
                            }
                        }
		    }	/* Make sure we have the rgb data and a positive geometry. */

                    /* Unref Imlib image, it is no longer needed. */
                    Imlib_destroy_image(imlib_handle, imlib_image);
                    imlib_image = NULL;

		}	/* Image loaded successfully? */
		else
		{
		    last_load_error = "Imlib was unable to load the image";
		}
	    }
#endif	/* HAVE_IMLIB */
#ifdef HAVE_IMLIB2
	    if(1)
	    {
		/* Load Imlib image, note that this places the
		 * image in cache if it was not already loaded or
		 * just increases the refcount for an image already
		 * loaded in the cache.
		 */
		imlib_image = imlib_load_image(filename);

		/* Image loaded successfully? */
		if(imlib_image != NULL)
		{
		    const u_int8_t *src_data_rgba;


                    status = 0;		/* Mark that load was successful. */

		    /* Set newly loaded Imlib image into Imlib2's
		     * context.
		     */
		    imlib_context_set_image(imlib_image);

		    /* Get size of newly loaded image. */
		    width = imlib_image_get_width();
		    height = imlib_image_get_height();
                    bpp = 4;
                    bpl = width * bpp;

		    /* Get pointer to Imlib image's RGBA data. */
		    src_data_rgba = (const u_int8_t *)imlib_image_get_data_for_reading_only();

		    /* Allocate return image data buffer. */
		    data_rgba = (u_int8_t *)realloc(
			data_rgba,
			bpl * height
		    );

		    /* Are source data, target data, and size all valid? */
		    if((src_data_rgba != NULL) && (data_rgba != NULL) &&
                       (width > 0) && (height > 0)
		    )
		    {
			memcpy(data_rgba, src_data_rgba, bpl * height);
		    }

		    /* Always deallocate Imlib2 images after
		     * rendering, this is efficient because the image
		     * is cached and managed by Imlib2.
		     * This will reduce the refcount of the image,
		     * when it is no longer needed then Imlib2 will
		     * actually deallocate it.
		     */
		    imlib_free_image();
		    imlib_image = NULL;

		}	/* Image loaded successfully? */
	    }
#endif	/* HAVE_IMLIB2  */
	}
	else if((io_type == IO_TYPE_PNG) && !user_aborted)
        {
#ifdef PNG_H
	    status = ImgPNGReadRGBA(
		filename,
		&width, &height, &bpl, &bpp,
		&data_rgba,
		bg_color,	/* 4 byte sin RGBA format, will be modified. */
		&creator, &author, &comments,
		def_alpha_value,
		client_data,
		progress_cb,
		&user_aborted
	    );
	    if((status != 0) && (status != -4))
	    {
		/* Do not update last_load_error, ImgPNGReadRGBA() should
		 * have updated it when an error was encountered.
		 */
/*		last_load_error = "???"; */
	    }
#endif  /* PNG_H */
	}
        else if((io_type == IO_TYPE_TGA) && !user_aborted)
        {
#ifdef TGA_H
            status = ImgTGAReadRGBA(
                filename,
                &width, &height, &bpl, &bpp,
                &data_rgba,
                bg_color,       /* 4 byte sin RGBA format, will be modified. */
                &creator, &author, &comments,
                def_alpha_value,
                client_data,
                progress_cb,
                &user_aborted
            );
            if((status != 0) && (status != -4))
            {
                /* Do not update last_load_error, ImgPNGReadRGBA() should
                 * have updated it when an error was encountered.
                 */
/*              last_load_error = "???"; */
            }
#endif  /* TGA_H */
        }


        /* Update returns. */
        *width_rtn = width;
        *height_rtn = height;
        *bpl_rtn = bpl;
        *bpp_rtn = bpp;
        *rgba_rtn = data_rgba;
	data_rgba = NULL;	/* Transfered to return, so set it NULL. */
        *creator_rtn = creator;
	creator = NULL;        /* Transfered to return, so set it NULL. */
        *author_rtn = author;
	author = NULL;        /* Transfered to return, so set it NULL. */
	*comments_rtn = comments;
	comments = NULL;	/* Transfered to return, so set it NULL. */

	/* If not user aborted, then call progress callback one last time. */
	if((progress_cb != NULL) && !user_aborted)
	{
            if(!progress_cb(
		client_data,
		1, 1,		/* Progress complete. */
		width, height,
		bpl, bpp,
		data_rgba
	    ))
                user_aborted = 1;
	}

	/* Do not deallocate data_rgba or comments, that is left to the
	 * calling function.
	 */

	if(user_aborted)
	{
            last_load_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(status);
}

/*
 *	Writes the image to file.
 *
 *      If progress_cb is not NULL then it will be called during loading,
 *      if it returns 0 then loading will be aborted.
 *
 *      Return values:
 *
 *	0	Success
 *	-1	General error
 *	-2	Unable to determine file format
 *	-3	Systems error
 *	-4	User abort
 */
int ImgWriteFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        const char *comments,   /* May be NULL. */
	float quality,		/* 0.0 to 1.0. */
	int save_as_color,	/* True means to save as color. */
        u_int8_t def_alpha_value,
	int imlib_fallback,	/* 0 = none or 1 = ImageMajick & NetPBM. */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
)
{
        int status;
        int user_aborted = 0;
        int io_type = IO_TYPE_ANY;
        const char *ext_ptr;


        /* Reset global last write error message pointer. */
        last_write_error = NULL;

        if((filename == NULL) || (rgba == NULL) ||
           (width < 1) || (height < 1) ||
           (bpp != 4)
        )
	{
	    last_write_error = "Invalid value used to describe image";
            return(-1);
	}

	/* Automatically calculate bytes per line? */
	if(bpl <= 0)
	    bpl = width * bpp;

        /* Get pointer to extension (if any) in file name. */
        ext_ptr = strrchr(filename, '.');
        if(ext_ptr == NULL)
            ext_ptr = "";

	/* Must have extension to determine format. */
	if((*ext_ptr) == '\0')
	{
            last_write_error = "Unable to determine format due to missing extension";
	    return(-2);
	}

	/* Explicitly set io type to use Imlib or Imlib2. */
	io_type = IO_TYPE_IMLIB;


        /* Reset status to -1 and load by IO type. */
        status = -1;

        /* Call progress callback for the first time as needed. */
        if(progress_cb != NULL)
        {
            if(!progress_cb(
                client_data,
                0, 1,           /* Progress starting. */
                width, height,
                bpl, bpp,
                rgba
            ))
                user_aborted = 1;
        }

        /* Use Imlib or Imlib2? */
        if((io_type == IO_TYPE_IMLIB) && !user_aborted)
        {
#ifdef HAVE_IMLIB
            if(imlib_handle != NULL)
            {
		int x, y, bc, min_rgb_bpp;
                ImlibImage *imlib_image = NULL;
                ImlibSaveInfo *si;

		int tbpl, tbpp;
		u_int8_t *tar_ptr, *tar_data_rgb, *tar_data_alpha;

		int sbpl, sbpp;
                const u_int8_t *src_ptr, *src_data;


		/* Get source image data. */
		sbpp = bpp;
		sbpl = bpl;
		src_data = (const u_int8_t *)rgba;

                /* Allocate target rgb and alpha buffers. */
                tbpp = 3;
                tbpl = width * tbpp;
                tar_data_rgb = (u_int8_t *)malloc(tbpl * height);
                tar_data_alpha = (u_int8_t *)malloc(width * height);
                if((tar_data_rgb != NULL) && (tar_data_alpha != NULL))
                {
                    /* Calculate smaller of rgb bytes per pixel. */
                    min_rgb_bpp = MIN(sbpp, tbpp);

                    /* Copy source image to our target buffers. */
                    for(y = 0; y < height; y++)
                    {
			if(progress_cb != NULL)
			{
			    if(!progress_cb(
				client_data,
				y, height,
				width, height,
				bpl, bpp,
				rgba
			    ))
			    {
				user_aborted = 1;
				break;
			    }
			}

                        for(x = 0; x < width; x++)
                        {
                            src_ptr = &src_data[
                                (y * sbpl) + (x * sbpp)
                            ];

                            /* Copy rgb. */
                            tar_ptr = &tar_data_rgb[
                                (y * tbpl) + (x * tbpp)
                            ];
                            for(bc = 0; bc < min_rgb_bpp; bc++)
                                *tar_ptr++ = *src_ptr++;

                            /* Copy alpha. */
                            tar_ptr = &tar_data_alpha[
                                (y * width) + (x * 1)
                            ];
                            /* More bytes left on source pixel pointer? */
                            if(bc < sbpp)
                            {
                                *tar_ptr = *src_ptr++;
                                bc++;
                            }
                            else
                            {
                                *tar_ptr = def_alpha_value;
                            }
                        }
                    }	/* Copy source image to our target buffers. */


                    /* Create Imlib image from loaded data only if
		     * user did not abort.
		     */
		    if(!user_aborted)
			imlib_image = Imlib_create_image_from_data(
			    imlib_handle,
			    (unsigned char *)tar_data_rgb,
			    (unsigned char *)tar_data_alpha,
			    width, height
			);
                    if(imlib_image != NULL)
                    {
                        /* Need to realize changes. */
                        Imlib_changed_image(imlib_handle, imlib_image);
                    }
		}
		/* Deallocate target buffers used to create the Imlib
		 * image. They are no longer needed.
		 */
		free(tar_data_rgb);
		tar_data_rgb = NULL;
		free(tar_data_alpha);
		tar_data_alpha = NULL;


                /* Allocate and set up save info. */
                si = (ImlibSaveInfo *)calloc(1, sizeof(ImlibSaveInfo));
                if(si != NULL)
                {
		    /* Quality for lossy formats (ie jpg), valid values
		     * from 0 = 0% to 256 = 100%.
		     */
		    si->quality = CLIP(quality, 0.0, 1.0) * 256;

		    /* Scaling for paper relative formats, valid values
		     * from 0 = 0% to 1024 = 100%.
		     */
		    si->scaling = 1024;

		    /* Bottom-left relative paper justification, valid
		     * values from 0 to 1024.
		     */
		    si->xjustification = 0;
		    si->yjustification = 1024;

		    si->page_size = PAGE_SIZE_LETTER;

		    /* Save as greyscale (value of 0) or color (value of 1). */
		    si->color = ((save_as_color) ? 1 : 0);
		}

                /* Set Imlib fallback:
		 *	0 = No fallback
		 *	1 = Use ImageMagick and NetPBM.
		 */
                Imlib_set_fallback(imlib_handle, imlib_fallback);

		/* Was the Imlib image created successfully? */
		if(imlib_image != NULL)
		{
		    char *dpath = strdup(filename);
		    /* Save Imlib image. */
		    status = Imlib_save_image(
			imlib_handle, imlib_image, dpath, si
		    );
		    free(dpath); dpath = NULL;

                    /* Save successful? */
                    if(status == 1)
                    {
			/* Update status to our interpritation of success. */
			status = 0;
		    }
		    else
		    {
			/* Update status to our interpritation of error. */
			status = -1;
		    }

                    /* Destroy Imlib image, it is no longer needed in
		     * the cache.
		     */
		    Imlib_kill_image(imlib_handle, imlib_image);
		    imlib_image = NULL;
		}
		/* At this point imlib_image has been deallocated. */

		/* Deallocate save info and any allocated members. */
		free(si);
		si = NULL;

	    }
	}
#endif	/* HAVE_IMLIB */
#ifdef HAVE_IMLIB2


#endif	/* HAVE_IMLIB2 */

        /* If not user aborted, then call progress callback one last time. */
        if((progress_cb != NULL) && !user_aborted)
        {
            if(!progress_cb(
                client_data,
                1, 1,           /* Progress complete. */
                width, height,
                bpl, bpp,
                rgba
            ))
                user_aborted = 1;
        }


        if(user_aborted)
	{
            last_write_error = "User aborted operation";
            return(-4);
	}
        else
            return(status);
}



/*
 *      Writes the PNG image to file using libpng.
 *
 *      If progress_cb is not NULL then it will be called during loading,
 *      if it returns 0 then loading will be aborted.
 *
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value
 *      -3      Systems error
 *      -4      User abort
 */
int ImgWritePNGFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format. */
	const char *creator, const char *author, const char *comments,
        int compression_level,  /* 0 to 9. */
        int interlaced,         /* 0 or 1. */
        int format,             /* 0 = b&w, 1 = grey, 2 = greyalpha
                                 * 3 = rgb, 4 = rgba.
                                 */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
)
{
#ifdef PNG_H
        int status;
        int user_aborted = 0;


        /* Reset global last write error message pointer. */
        last_write_error = NULL;

        if((filename == NULL) || (rgba == NULL) ||
           (width < 1) || (height < 1) ||
           (bpp != 4)
        )
        {
            last_write_error = "Invalid value used to describe image";
            return(-1);
        }

        /* Automatically calculate bytes per line? */
        if(bpl <= 0)
            bpl = width * bpp;

        /* Call progress callback for the first time as needed. */
        if(progress_cb != NULL)
        {
            if(!progress_cb(
                client_data,
                0, 1,           /* Progress starting. */
                width, height,
                bpl, bpp,
                rgba
            ))
                user_aborted = 1;
        }

	/* Call PNG writing function for RGBA image data format. */
	status = ImgPNGWriteRGBA(
	    filename,
	    width, height, bpl, bpp,
	    rgba,
	    bg_color,
	    creator, author, comments,
	    compression_level,
	    interlaced,
	    format,		/* 0 = b&w, 1 = grey, 2 = greyalpha
                                 * 3 = rgb, 4 = rgba.
                                 */
	    client_data, progress_cb,
	    &user_aborted
	);


        /* If not user aborted, then call progress callback one last time. */
        if((progress_cb != NULL) && !user_aborted)
        {
            if(!progress_cb(
                client_data,
                1, 1,           /* Progress complete. */
                width, height,
                bpl, bpp,
                rgba
            ))
                user_aborted = 1;
        }


        if(user_aborted)
        {
            last_write_error = "User aborted operation";
            return(-4);
        }
        else
            return(status);
#else
	/* PNG_H is not defined suggesting that libpng is not available,
	 * so instead just call ImgWriteFileRGBA().
	 */
	return(
	    ImgWriteFileRGBA(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		comments,
		1.0,		/* Quality. */
		1,		/* Save as color. */
		0xff,		/* Default alpha value. */
		1,		/* Imlib fallback. */
		client_data,
		progress_cb
	    )
	);
#endif  /* PNG_H */
}

/*
 *      Writes the TGA image to file using the tga library.
 *
 *      If progress_cb is not NULL then it will be called during
 *	loading, if it returns 0 then loading will be aborted.
 *
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value
 *      -3      Systems error
 *      -4      User abort
 */
int ImgWriteTGAFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        int format,             /* 0 = greyscale, 1 = rgb, 2 = rgba */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
)
{
#ifdef TGA_H
        int status;
        int user_aborted = 0;


        /* Reset global last write error message pointer. */
        last_write_error = NULL;

        if((filename == NULL) || (rgba == NULL) ||
           (width < 1) || (height < 1) ||
           (bpp != 4)
        )
        {
            last_write_error = "Invalid value used to describe image";
            return(-1);
        }

        /* Automatically calculate bytes per line? */
        if(bpl <= 0)
            bpl = width * bpp;

        /* Call TGA writing function for RGBA image data format. */
        status = ImgTGAWriteRGBA(
            filename,
            width, height, bpl, bpp,
            rgba,
            bg_color,
	    format,
            client_data, progress_cb,
            &user_aborted
        );

        /* If not user aborted, then call progress callback one last
	 * time.
	 */
        if((progress_cb != NULL) && !user_aborted)
        {
            if(!progress_cb(
                client_data,
                1, 1,           /* Progress complete. */
                width, height,
                bpl, bpp,
                rgba
            ))
                user_aborted = 1;
        }

        if(user_aborted)
        {
            last_write_error = "User aborted operation";
            return(-4);
        }
        else
            return(status);

#else
	/* TGA_H is not defined suggesting that tga library is not
	 * available, so instead just call ImgWriteFileRGBA().
	 */
        return(
            ImgWriteFileRGBA(
                filename,
                width, height, bpl, bpp,
                rgba,
                bg_color,
                comments,
                1.0,            /* Quality. */
                1,              /* Save as color. */
                0xff,           /* Default alpha value. */
                1,              /* Imlib fallback. */
                client_data,
                progress_cb
            )
        );
#endif  /* TGA_H */
}

