#ifndef IMGIO_H
#define IMGIO_H

#include <sys/types.h>

extern const char *ImgLoadGetError(void);
extern const char *ImgWriteGetError(void);
extern int ImgLoadFileRGBA(
	const char *filename,
	int *width_rtn, int *height_rtn,
	int *bpl_rtn, int *bpp_rtn,	/* bpp_rtn will always be 4. */
	u_int8_t **rgba_rtn,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format, will be modified. */
	char **creator_rtn, char **author_rtn, char **comments_rtn,
	u_int8_t def_alpha_value,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data. */
		int, int,	/* Progress min and max. */
		int, int,	/* Width and height. */
		int, int,	/* Bytes per line and bytes per pixel. */
		u_int8_t *	/* RGBA data. */
	)
);
extern int ImgWriteFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,	/* 4 bytes in RGBA format. */
        const char *comments,   /* May be NULL. */
        float quality,          /* 0.0 to 1.0. */
        int save_as_color,      /* True means to save as color. */
        u_int8_t def_alpha_value,
        int imlib_fallback,     /* 0 = none or 1 = ImageMajick & NetPBM. */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
);
extern int ImgWritePNGFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        const char *creator, const char *author, const char *comments,
        int compression_level,  /* 0 to 9. */
        int interlaced,         /* 0 or 1. */
        int format,		/* 0 = b&w, 1 = grey, 2 = greyalpha
                                 * 3 = rgb, 4 = rgba.
                                 */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
);
extern int ImgWriteTGAFileRGBA(
        const char *filename,
        int width, int height,
        int bpl, int bpp,       /* bpp must be 4. */
        u_int8_t *rgba,
        u_int8_t *bg_color,     /* 4 bytes in RGBA format. */
        int format,             /* 0 = greyscale, 1 = rgb, 2 = rgba */
        void *client_data,
        int (*progress_cb)(
                void *,         /* Client data. */
                int, int,       /* Progress min and max. */
                int, int,       /* Width and height. */
                int, int,       /* Bytes per line and bytes per pixel. */
                u_int8_t *      /* RGBA data. */
        )
);

#endif	/* IMGIO_H */
